/*------------------------------------------------------------------*-C-*-*
 * %Z%1.12  %G% 18:30:08 %W%
 *
 * Purpose:	
 *
 *------------------------------------------------------------------------*
 * Notes:
 *------------------------------------------------------------------------*/

char *__rstore_util_sccsid__() { return "%Z%%W% 1.12 %G% 18:30:08"; }

#include <stdio.h>
#include <string.h>
#include "rstoret.h"
#include "alloc.h"
#include "mmglue.h"
#include <rscheme/scheme.h>
#include <rscheme/writebar.h>
#include "indirect.h"
#include "swizzler.h"

/* usage of lss's client misc[] area */

struct swiz_mode_handler *(app_swiz_mode_handlers[10]);

extern IRC_Heap *gc_arena;

static RStore *(rstores[10]);
static unsigned num_rstores;
obj alloc_area_class; /* initialized by (open-rstore) */

void rstore_error_handler( LSS *lss, void *info, int code, char *msg )
{
  scheme_error( "lss error: ~a (~d)\nin ~s",
	        3,
	        make_string(msg),
	        int2fx(code), 
	        ((RStore *)info)->owner );
}

void plain_lss_error_handler( LSS *lss, void *info, int code, char *msg )
{
  scheme_error( "lss error: ~a (~d)",
	        2,
	        make_string(msg),
	        int2fx(code) );
}

struct swiz_mode_handler *get_swiz_mode_handler( RStore *sto,
						 enum SwizzleMode mode )
{
  struct swiz_mode_handler **p;

  for (p=sto->swiz_mode_handlers; *p; p++)
    {
      if ((*p)->handles == mode)
	return *p;
    }
  scheme_error( "image mode ~d not handled by ~s", 
		2, int2fx(mode), sto->owner );
  return NULL;
}

void rstore_add_swiz_mode_handler( RStore *sto, struct swiz_mode_handler *h )
{
  struct swiz_mode_handler **p;

  if (h->handles >= SWIZ_MODE_APP_0 && h->handles <= SWIZ_MODE_APP_9)
    {
      for (p=sto->swiz_mode_handlers; *p; p++)
	{
	  if ((*p)->handles == h->handles)
	    {
	      scheme_error( "swizzle mode ~d already handled", 
			    1, int2fx( h->handles ) );
	    }
	}
      if (p >= (sto->swiz_mode_handlers + 10))
	scheme_error( "swizzle mode handler table full", 0 );
      *p++ = h;
      *p = NULL;
    }
  else
    scheme_error( "swizzle mode ~d not valid for app swizzler",
		  1, int2fx( h->handles ) );
}

struct RStore *rstore_open( obj owner, const char *lss_path, int lss_mode, off_t CR_offset )
{
struct RStore *store;
LSS *lss;
static rs_bool mminit = NO;
struct VMPageRecord *vmp0;

    if (!mminit)
      {
	unsigned i;

	init_mm();
	mminit = YES;
      }

    store = ALLOC( struct RStore );
    store->owner = owner;

    lss = lss_open( lss_path, lss_mode, 
		    rstore_error_handler,
		    store,
		    CR_offset );
    if (!lss)
	return NULL;

    rstores[num_rstores++] = store;

    store->swiz_mode_handlers[0] = NULL;
    store->lss = lss;
    store->pivot_table = FALSE_OBJ;
    store->local_code_ptrs = FALSE_OBJ;
    store->local_fn_descrs = FALSE_OBJ;

    /* initialize our fake size class and generation */

    store->the_gen.heap = gc_arena;

    store->the_size_class.gen = &store->the_gen;
    store->the_size_class.heap = gc_arena;

    /* install write-barrier hooks */

    {
      unsigned l, r;

      l = 0xF;
      for (r=0; r<14; r++)
	gc_arena->writeBarrierTable[l*16+r] = WB_PERSISTENT;
    }
      

    /* empty the lists of various VMPs */

    store->first_dirty = NULL;
    store->last_dirty = NULL;
    store->num_dirty = 0;
    store->first_loaded = NULL;
    store->first_reserved = NULL;

    htable_init( &store->vm_page_records );
    htable_init( &store->reserved_base_pages );
    
    if (lss_mode & LSS_CREATE)
    {
      struct PHeapHdr *phh;

      store->next_page = 0x1000000;
      store->next_indirect_page = 256;   /* 0...63 reserved for the system,
					    64..255 for app. pivot pages */

      /* create the first allocation area */

      store->default_area = make_alloc_area( store, FALSE_OBJ );
    }
    else
    {
      struct PageRef pr0;
      struct RStoreCommitInfo *ci;
      
      ci = lss_get_client_commit_info( store->lss, NULL );
      
      store->next_page = ci->next_page_num;
      store->next_indirect_page = ci->next_indirect_page;
      store->default_area = PTR_TO_DATAPTR(translate_LR( store,
							 ci->default_area ));
    }
    return store;
}

/* allocate some new storage... */

void install_new_vmpr( RStore *store, struct VMPageRecord *vmpr )
{
    vmpr->next_reserved = store->first_reserved;
    store->first_reserved = vmpr;

    if (vmpr->ref.loaded)
      {
	vmpr->next_loaded = store->first_loaded;
	store->first_loaded = vmpr;
      }

    if (vmpr->ref.dirty)
      {
	if (store->first_dirty)
	  {
	    /* put this after the previously last dirty page */
	    store->last_dirty->next_dirty = vmpr;
	  }
	else
	  {
	    /* this is the first dirty page... */
	    store->first_dirty = vmpr;
	  }
	store->last_dirty = vmpr;
	vmpr->next_dirty = NULL;
	store->num_dirty++;
      }

    if (!vmpr->ref.indirect)
      htable_insert( &store->vm_page_records, vmpr->mem_address )->value = vmpr;
    if (vmpr->ref.first)
      htable_insert( &store->reserved_base_pages,
		    (void *)vmpr->ref.base_page_num )->value = vmpr;
}


struct VMPageRecord *alloc_ppages( RStore *store,
				   UINT_32 base_page_num, UINT_32 num_pages )
{
struct VMPageRecord *first, *vmpr;
void *addr;
unsigned i;

  /* allocate the first */

    addr = mm_alloc( num_pages * MM_PAGE_SIZE, MM_MODE_READ_WRITE );

    first = vmpr = ALLOC( struct VMPageRecord );

    vmpr->ref.base_page_num = base_page_num;
    vmpr->ref.first = 1;
    vmpr->ref.indirect = 0;
    vmpr->ref.dirty = 1;
    vmpr->ref.loaded = 1;
    vmpr->ref.nth_page = num_pages;  /* for FIRST, is # pages */
    vmpr->mem_address = addr;

    install_new_vmpr( store, vmpr );
    
    /* allocate the rest */

    for (i=1; i<num_pages; i++) 
      {
	vmpr = ALLOC( struct VMPageRecord );

	vmpr->ref.base_page_num = base_page_num;
	vmpr->ref.first = 0;
	vmpr->ref.indirect = 0;
	vmpr->ref.dirty = 1;
	vmpr->ref.loaded = 1;
	vmpr->ref.nth_page = i;
	addr = ((char *)addr) + MM_PAGE_SIZE;
	vmpr->mem_address = addr;
	
	install_new_vmpr( store, vmpr );
      }
    return first;
}

/* reserve a new page (or pages) */

static struct VMPageRecord *reservation( RStore *store, 
					 struct PageRef *pr, void *addr )
{
struct VMPageRecord *vmpr = ALLOC( struct VMPageRecord );

    vmpr->ref = *pr;
    vmpr->ref.dirty = 0;
    vmpr->ref.loaded = 0;
    
    vmpr->next_reserved = store->first_reserved;
    store->first_reserved = vmpr;

    vmpr->next_dirty = NULL;
    vmpr->next_loaded = NULL;

    vmpr->mem_address = addr;
    
    htable_insert( &store->vm_page_records, addr )->value = vmpr;
    if (pr->first)
      {
	htable_insert( &store->reserved_base_pages,
		       (void *)pr->base_page_num )->value = vmpr;
      }
    return vmpr;
}

struct VMPageRecord *reserve_multi_page( struct RStore *store, 
					 struct PageRef *pr )
{
  unsigned i, n;
  char *a;
  struct VMPageRecord *first;

  assert( pr->first ); /* only works if we encounter the 1st page! */

  n = pr->nth_page;
  a = (char *)mm_alloc( MM_PAGE_SIZE * n, MM_MODE_NO_ACCESS );

  first = reservation( store, pr, a );

  for (i=1; i<n; i++)
    {
      struct PageRef interior;

      interior.base_page_num = pr->base_page_num;
      interior.first = 0;
      interior.indirect = 0;
      interior.dirty = 0;
      interior.loaded = 0;
      interior.nth_page = i;

      a += MM_PAGE_SIZE;
      reservation( store, &interior, a );
    }
  return first;
}

struct VMPageRecord *reserve_single_page( struct RStore *store, 
					  struct PageRef *pr )
{
  return reservation( store, pr, mm_alloc( MM_PAGE_SIZE, MM_MODE_NO_ACCESS ) );
}

/* returns an existing vmpr, or reserves it */

struct VMPageRecord *get_vmpr( struct RStore *store, struct PageRef *pr )
{
struct htent *e;

    e = htable_lookup( &store->reserved_base_pages,
		       (void *)pr->base_page_num );
    if (e)
	return (struct VMPageRecord *)e->value;

    if (pr->indirect)
	return build_indirect_page( store, pr );
    else if (pr->first && pr->nth_page == 1)
	return reserve_single_page( store, pr );
    else
	return reserve_multi_page( store, pr );
}

void rstore_close( struct RStore *store )
{
  lss_close( store->lss );
}

static obj attempt_to_flush_pages( struct RStore *store, obj root_obj )
{
  if (!EQ(root_obj,store->default_area->entry))
    {
      gvec_set( DATAPTR_TO_PTR(store->default_area),
	        SLOT(0),
	        root_obj );
    }
  return write_dirty_pages( store );
}

static obj do_commit( struct RStore *store, obj root )
{
  struct RStoreCommitInfo ci;
  unsigned i;
  int rc;
  
  memset( &ci, 0, sizeof ci );
  strcpy( ci.tag, "TKG-RStore-1.0" );
  
  ci.next_page_num = store->next_page;
  ci.next_indirect_page = store->next_indirect_page;
  ci.default_area = create_LR_first( store,
				     DATAPTR_TO_PTR(store->default_area) );
    
  lss_commit( store->lss, &ci, sizeof ci );

  return NIL_OBJ;
}

obj rstore_commit( struct RStore *store, obj root )
{
obj unresolved = attempt_to_flush_pages( store, root );

    if (EQ(unresolved,NIL_OBJ))
      return do_commit( store, root );
    return unresolved;
}

void dirty_page( RStore *store, struct VMPageRecord *vmp )
{
    assert( !vmp->ref.dirty );
    if (store->first_dirty)
      {
	store->last_dirty->next_dirty = vmp;
      }
    else
      {
	store->first_dirty = vmp;
      }
    store->last_dirty = vmp;
    store->num_dirty++;

    vmp->next_dirty = NULL;
    mm_set_prot( vmp->mem_address, MM_PAGE_SIZE, MM_MODE_READ_WRITE );
}

void mmc_access_failed( void *addr )
{
  struct VMPageRecord *vmp;
  RStore *store;
  char temp[12];

#ifdef VERBOSE
  printf( "protection fault at: %p\n", addr );
#endif

  /* look up the address */
  vmp = find_owner_and_vmpr( addr, &store );

  if (vmp)
    {
      if (!vmp->ref.loaded)
	{
	  /* load it in */
	  load_page( store, vmp );
	}
      else 
	{
	  /* mark it dirty */
	  assert( !vmp->ref.dirty );
	  dirty_page( store, vmp );
	}
    }
  else
    {
      sprintf( temp, "%08x", (UINT_32)addr );
      scheme_error( "access protection error at address #x~a",
		   1, make_string(temp) );
    }
}
 
struct VMPageRecord *find_owner_and_vmpr( void *addr, RStore **s )
{
  int i;
  struct VMPageRecord *vmp;

  for (i=0; i<num_rstores; i++)
    {
      vmp = addr_to_vm_page_record( rstores[i],
				    MM_PAGE_BASE_ADDR(addr) );
      if (vmp)
	{
	  *s = rstores[i];
	  return vmp;
	}
    }
  return NULL;
}


enum SwizzleMode mode_for_object( struct PHeapHdr *hdr )
{
  obj imode = gvec_read( hdr->rs_header.pob_class, SLOT(2) );
  enum SwizzleMode m;

  m = OBJ_ISA_FIXNUM(imode) ? fx2int(imode) : SWIZ_MODE_GVEC;
  switch (m)
    {
    case SWIZ_MODE_GVEC:
    case SWIZ_MODE_BVEC:
    case SWIZ_MODE_TEMPLATE:
    case SWIZ_MODE_UINT32:
    case SWIZ_MODE_FLOAT:
    case SWIZ_MODE_PADDR_VEC:
    case SWIZ_MODE_PART_DESCR:
    case SWIZ_MODE_ALLOC_AREA:
    case SWIZ_MODE_APP_0:
    case SWIZ_MODE_APP_1:
    case SWIZ_MODE_APP_2:
    case SWIZ_MODE_APP_3:
    case SWIZ_MODE_APP_4:
    case SWIZ_MODE_APP_5:
    case SWIZ_MODE_APP_6:
    case SWIZ_MODE_APP_7:
    case SWIZ_MODE_APP_8:
    case SWIZ_MODE_APP_9:
      return m;
    }
  scheme_error( "don't know how to swizzle ~s (image-mode ~s)",
	        2, PHH_TO_PTR(hdr), imode );
  return 0; /* hush up -Wall, since scheme_error() never returns */
}

void rstore_copy_gvec_proc( obj new_item, obj old_item )
{
  UINT_32 i, size = PTR_TO_HDRPTR(old_item)->pob_size;

  for (i=0; i<size; i+=SLOT(1))
    {
      gvec_write_init( new_item, i, gvec_ref( old_item, i ) );
    }
}

/* copy an object into the store */

obj copy_into_pstore( PAllocArea *area, obj item )
{
  obj new_item;
  POBHeader *h;
  enum SwizzleMode m;

  if (!OBJ_ISA_PTR(item))
    {
      /* silently ignore requests to copy non-ptrs into the store */
      return item;
    }

  h = PTR_TO_HDRPTR(item);

  m = mode_for_object(PTR_TO_PHH(item));

  if (m != SWIZ_MODE_ALLOC_AREA)
    new_item = parea_alloc( (AllocArea *)area, h->pob_class, h->pob_size );
  else
    new_item = FALSE_OBJ;

  switch (m)
    {
    case SWIZ_MODE_BVEC:
    case SWIZ_MODE_UINT32:
    case SWIZ_MODE_FLOAT:
    case SWIZ_MODE_PADDR_VEC:
      memcpy( PTR_TO_DATAPTR(new_item), 
	     PTR_TO_DATAPTR(item), 
	     h->pob_size );
      break;
    case SWIZ_MODE_ALLOC_AREA:  /* quiet compiler */
      scheme_error( "Illegal copy: ~s", 1, item );
    case SWIZ_MODE_GVEC:
    case SWIZ_MODE_PART_DESCR:
    case SWIZ_MODE_TEMPLATE:
      rstore_copy_gvec_proc( new_item, item );
      break;
    default:
      {
	struct swiz_mode_handler *h = get_swiz_mode_handler( area->owner, m );
	h->copy_in( new_item, item );
      }
    }
  return new_item;
}

/* returns an assoc-list of all module names & their module ptrs */

obj all_modules( void )
{
  obj lst = NIL_OBJ;
  struct module_descr **p;

  for (p=master_table; *p; p++)
    {
      lst = cons( cons( make_string( (*p)->name ),
		        RAW_PTR_TO_OBJ( *p ) ),
		  lst );
    }
  return lst;
}
